/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "oirSupport.h"
#include "fitsio.h"
#include "imaging_fdata.h"
#include "minrtsFftUtil.h"
#define NBIN 5
#define maxshift 4
/**********************************************************
 * Usage:
 *    oirMakePhotoSpectra -A AFile -B BFile -AB ABFile -mask maskFile 
 *       -out outFile [-shift maskShift] [-autoshift] [-outmask outMaskFile]
 *
 *    AFile and BFile are the output of oirChopPhotoImages for AOPEN and
 *    BOPEN shutters respectively
 *
 *    maskFile defines the pixel weighting before collapsing the spectra in Y
 *
 *    -shift asks the program to move the mask up/down maskShift pixels before
 *    masking; -autoshift figures out maskShift itself (and overrides -shift)
 *
 *    -outmask asks to program to output the mask calculated on
 *    the basis of maskShift or -autoshift so that you can use
 *    it for other purposes (like estimating correlated fluxes).
 *
 *    OutFile is a IMAGING_DATA table in a fits file with
 *    12-rows, each with only one DATA array (sum of channels 1 and 2)
 *
 *    1:  Total Flux A-side (ADUs/s/channel)
 *    2:  Total Flux B-side 
 *    3:  Total Flux Geometric Mean AB
 *    4:  Masked Flux A-side 
 *    5:  Masked Flux B-side 
 *    6:  Masked Flux Geometric Mean AB
 *    7->12 are estimated RMS errors in rows 1->6 resp.
 *
 ************************************************************* */

float minrtsqpk (int nData, float *x, float *y, float *ymax, int domin) {
/* make quadratic fit to peak of y(x); return position of peak
   and value of peak */
   int i, j;
   float maxval, jmax, xm;
   maxval = y[0];
   j   = 0;
   /* find maximum or maximum, as requrested */
   if (domin == 0) {
      for (i=0; i<nData; i++) {
         if (y[i] > maxval) {
            maxval = y[i];
            j = i;
         }
      }
   }  /* integral pixel position of peak */
   else {
      for (i=0; i<nData; i++) {
         if (y[i] < maxval) {
            maxval = y[i];
            j   = i;  
         }
      } 
   }  /* integral pixel position of peak */
/*   j = vmax(j, 2);
   j = vmin(j, nData-3); */
   jmax = -0.5*(y[j+1]-y[j-1])/(y[j+1]+y[j-1]-2*y[j]); /* fractional pixel position */
   xm = x[j] + .5*(x[j+1]-x[j-1])*jmax +
      .5*(x[j+1]+x[j-1]-2*x[j])*jmax*jmax; /* quad interpolated position of pk */
   *ymax = y[j] + .5*(y[j+1]-y[j-1])*jmax +
      .5*(y[j+1]+y[j-1]-2*y[j])*jmax*jmax; /* quad interpolated value of pk */
   return xm; 
} /* end of minrtsQPk */
void Usage() {
   printf("Usage:\n");
   printf("  oirMakePhotoSpectra -A Afile -B Bfile -AB ABfile -mask maskFile -out outFile\n");
   printf("     [-shift maskShift] [-autoshift] [-outmask outMaskFile]\n\n");
   printf("     AFile contains the output of oirChopPhotoImages for AOPEN shutter\n");
   printf("     BFile contains the output of oirChopPhotoImages for BOPEN shutter\n");
   printf("     ABFile contains the output of oirChopPhotoImages for SCIPHOT ABOPEN shutter\n");
   printf("     maskFile defines pixel weighting before collapsing in Y-direction\n");
   printf("     outFile defines output file\n");
   printf("     maskShift [optional] shifts mask up (if +) or down (-) given # of pixels \n");
   printf("        (can be fractional pixels)\n");
   printf("     autoshift determines maskShift by crosscorrelating mask and data\n");
   printf("        outMaskFile defines where to (optionally) output new mask after\n");
   printf("     maskShift/automask operations.\n"); 
   printf("     \nn");
   return;
}
/************************************************************
 *     determine best shift of mask by finding maximum signal
 *     empirically
 *******************************************************************/
   float findMaskShift(float *mask1, float *mask2, 
      float *A1, float *A2, float *B1, float *B2, int nx, int ny) {
      float Acorr[2*maxshift+1], Bcorr[2*maxshift+1];
      float shift[2*maxshift+1];
      double dA, dB;
      float x;
      int i,ix, iy, iShift, is;
      long xy, sy;
      for (iShift=(-maxshift);iShift<=maxshift;iShift++) {
         is = iShift + maxshift;
         shift[is] = iShift;
         Acorr[is] = 0.;
         Bcorr[is] = 0.;
         for (iy=maxshift;iy<(ny-maxshift);iy++) {
            xy = iy*nx;
            sy = (iy + iShift)*nx;
            for (ix=0;ix<nx;ix++) {
               Acorr[is] = Acorr[is] 
                  + A1[ix+xy]*mask1[ix+sy] 
                  + A2[ix+xy]*mask2[ix+sy] ;
               Bcorr[is] = Bcorr[is] 
                  + B1[ix+xy]*mask1[ix+sy] 
                  + B2[ix+xy]*mask2[ix+sy];
            }
         } /* yloop */
      } /* iShift loop */
      dA = minrtsQPk(2*maxshift+1, shift, Acorr, &x, 0);
      dB = minrtsQPk(2*maxshift+1, shift, Bcorr, &x, 0);
      printf(" For A and B data, shift peaks are %f %f\n",dA, dB);
      dA = -.5*(dA+dB);
      return dA;
   }  /* end findMaskShift
/**********************************************************
   shift each of the nRegion masks up (if shift >0) or down the given number of
   pixels.  
***********************************************************/

void shiftMaskImages(float** maskImages, float shift, int nRegion, int nx, int ny) {
   int iRegion, ix, iy, oy;
   float *oldy, *newy;
   float *olddata, *newdata;

   oldy    = (float *)malloc(ny * sizeof(float));
   newy    = (float *)malloc(ny * sizeof(float));
   olddata = (float *)malloc(ny * sizeof(float));

   if (shift > (ny-3)) shift = ny-3;
   if (shift < (3-ny)) shift = 3-ny;
   for (iy=0;iy<ny;iy++) {
      oldy[iy] = (float)iy;
      newy[iy] = oldy[iy] - shift;
   }
   if (shift == 0.) return;
   for (iRegion=0;iRegion<nRegion; iRegion++) {
      for (ix=0;ix<nx;ix++) {
         for(iy=0;iy<ny;iy++) olddata[iy] = maskImages[iRegion][ix+nx*iy];
         newdata = minrtsSpline(ny, oldy, olddata, ny, newy, .1);
         for(iy=0;iy<ny;iy++) maskImages[iRegion][ix+nx*iy] = newdata[iy];
         free(newdata);
      }
   }  /* iRegion */
   free(oldy);
   free(newy);
   free(olddata);
   return;
}   /* end of shift MaskImages */
void putOutMask(float **maskImages, int nRegion, int nx, int ny, 
   char *outMaskFileName, imaging_detector_table_def *DetectorDef, int *ierr) {

   fitsfile *outMaskFile;
   imaging_fdata_row *maskData ;
   imaging_fdata_table_def *mdata;
   int iRegion, ix, iy;

   *ierr = 0;
   fits_create_file (&outMaskFile, outMaskFileName, ierr);
   if (*ierr !=0) {
      fprintf (stderr, "error opening mask output %s %i\n",outMaskFileName, *ierr);
      printf("%s \n",strerror(errno));
      return ;
   }
   mdata = make_imaging_fdata_table_def (outMaskFile, DetectorDef, 1, ierr);
   if (*ierr !=0) {
      fprintf (stderr, "error creating mask data definition\n");
      return ;
   }
   maskData = make_imaging_fdata_row(mdata);
   for (iRegion=0;iRegion<nRegion;iRegion++)
      for (ix=0;ix<nx;ix++) for (iy=0;iy<ny;iy++)
         maskData->data[iRegion][ix+iy*nx] = maskImages[iRegion][ix+iy*nx];
   ix = write_imaging_fdata_table(outMaskFile, mdata, maskData, 1, ierr);
   fits_close_file(outMaskFile, ierr);
   mdata->detector=NULL;
   kill_imaging_fdata_table_def(mdata);
   kill_imaging_fdata_row(maskData);
   return;
}

int main (int argc, char** argv) {

   int i, j, ierr;
   int chan1;
   int nx, ny;
   long npix;
   long iPix;
   int  ix, iy;
   float xx, yy, mm;
   long  ii;
   int iKey;
   long nRow;
   int iarg, iBin;

   char outFileName[130], outMaskFileName[130];
   char *AFile;
   char *BFile;
   char *ABFile;
   char *maskFile;


   imaging_fdata_table_def *odata, *idata;
   imaging_detector_table_def *DetectorDef;
   imaging_fdata_row *outData, *inData;
   float *maskImages[12];
   int nxList[12], nyList[12];
   int nRegion;

   float *ATarget[2];
   float *BTarget[2];
   float *ABTarget[2];
   float *geoMean[2];
   float *ARms[2], *BRms[2];

   float **ABin[2], **BBin[2];
   float **ABgin[2];
   float Ascratch, Bscratch, ABscratch;

   float *AMask, *BMask, *ABMask;
   float *ATotal, *BTotal, *ABTotal;

   float *ARMask, *BRMask, *ABRMask;
   float *ARTotal, *BRTotal, *ABRTotal;

   int   nbin;
   char  opt1[12];

   fitsfile *inFile;
   fitsfile *outFile = NULL;

   char *highSens = "HIGH_SENS";
   char *bang="!";
   int yTop    = 25;
   int yBottom = 5;
   int one     = 1;
   int doAB    = 0;
   int autoShift      = 0;
   float maskShift    = 0;
   outMaskFileName[0] = 0;
   ABFile             = NULL;
   ierr = 0;

   printf("\n      ****** oirMakePhotoSpectra ****** \n");
/*
      parse input
*/
   iarg  = 1;
   if (argc <2 ){
      Usage();
      return 1;
   }
   while(iarg < argc ) {
      if(!strcmp("-A", argv[iarg])) AFile = argv[++iarg];
      if(!strcmp("-B", argv[iarg])) BFile = argv[++iarg];
      if(!strcmp("-AB", argv[iarg])) ABFile = argv[++iarg];
      if(!strcmp("-mask", argv[iarg])) maskFile = argv[++iarg];
      if(!strcmp("-shift", argv[iarg])) sscanf(argv[++iarg], "%f", &maskShift);
      if(!strcmp("-autoshift", argv[iarg])) autoShift = 1;
      if(!strcmp("-out", argv[iarg])) { 
         strcpy(outFileName, bang);
         strcat(outFileName, argv[++iarg]);
      }
      if(!strcmp("-outmask", argv[iarg])) { 
         strcpy(outMaskFileName, bang);
         strcat(outMaskFileName, argv[++iarg]);
      }
      ++iarg;
   }
   if (AFile[0]==0) {
      printf("You must specify an A input file!\n");
      Usage();
      return 1;
   }
   if (BFile[0]==0) {
      printf("You must specify an B input file!\n");
      Usage();
      return 1;
   }
   if (outFileName[0]==0) {
      printf("You must specify an output file!\n");
      Usage();
      return 1;
   } else printf("   Writing output photometry to %s.\n", outFileName);
   if (ABFile != NULL) {
      printf(" Using ABFile %s \n", ABFile);
      doAB = 1;
   }
   if (autoShift) printf("Determining maskShift by crosscorrelation.\n");
   if (outMaskFileName[0]==0) 
      printf("   Not saving output mask file.\n");
   else
      printf("Saving actual maskFile in %s.\n", outMaskFileName);
/*
      open Afile and get a lot of parameters for iShutter=1
*/
   i = fits_open_file(&inFile, AFile, READONLY, &ierr) ;
   if (ierr != 0) {
      printf("cfitsio error %i opening A file %s\n", ierr, AFile);
      printf("%s \n",strerror(errno));
      return 1;
   }
   ierr = 0;
   i = fits_read_key(inFile, TSTRING, "HIERARCH ESO INS OPT1 NAME", opt1,0, &ierr);
   printf("   beam combiner                = %s\n", opt1);
   if (strncmp(opt1, highSens, 9) == 0) chan1 = 0; else chan1 = 1;
   i = fits_read_key(inFile, TINT, "NBIN", &nbin,0, &ierr);
   printf("   NBIN                         = %i\n",nbin);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening fits output %s %i\n",outFileName, ierr);
      printf("%s \n",strerror(errno));
      return 1;
   }
  i = fits_copy_header(inFile, outFile, &ierr);
/*
      move to detector section
*/
   DetectorDef = read_imaging_detector_table(inFile, &ierr);
   if (ierr !=0) {
      printf("error %i opening detector table in %s\n", 
         ierr, AFile);
      return 1;
   }
   nRegion = DetectorDef->nregion;
   for (i=0;i<nRegion;i++) {
      nxList[i] = DetectorDef->naxis[i][0];
      nyList[i] = DetectorDef->naxis[i][1];
   }
   nx = DetectorDef->naxis[0][0];
   ny = DetectorDef->naxis[0][1];

/*
      allocate accumulator arrays
*/

   npix = ny * (long)nx;
   for (j=0;j<2;j++) {
      ATarget[j] = (float *)malloc(npix * sizeof(float));
      memset(ATarget[j], 0, npix*sizeof(float));
      BTarget[j] = (float *)malloc(npix * sizeof(float));
      memset(BTarget[j], 0, npix*sizeof(float));
      ABin[j]    = (float **)malloc(nbin * sizeof(float *));
      BBin[j]    = (float **)malloc(nbin * sizeof(float *));
      for (i=0; i<nbin; i++) {
         ABin[j][i] = (float *)malloc(npix * sizeof(float));
         memset(ABin[j][i], 0, npix*sizeof(float));
         BBin[j][i] = (float *)malloc(npix * sizeof(float));
         memset(BBin[j][i], 0, npix*sizeof(float));
      }
      if (doAB) {
         ABTarget[j] = (float *)malloc(npix * sizeof(float));
         memset(ABTarget[j], 0, npix*sizeof(float));
         ABgin[j]    = (float **)malloc(nbin * sizeof(float *));
        for (i=0; i<nbin; i++) 
           ABgin[j][i] = (float *)malloc(npix * sizeof(float));
           memset(ABgin[j][i], 0, npix*sizeof(float));
     
      }
   }
   AMask  = (float *)malloc(nx * sizeof(float));
   memset(AMask, 0, nx*sizeof(float));
   BMask  = (float *)malloc(nx * sizeof(float));
   memset(BMask, 0, nx*sizeof(float));
   ABMask = (float *)malloc(nx * sizeof(float));
   memset(ABMask, 0, nx*sizeof(float));
   ATotal  = (float *)malloc(nx * sizeof(float));
   memset(ATotal, 0, nx*sizeof(float));
   BTotal  = (float *)malloc(nx * sizeof(float));
   memset(BTotal, 0, nx*sizeof(float));
   ABTotal = (float *)malloc(nx * sizeof(float));
   memset(ABTotal, 0, nx*sizeof(float));

   ARMask  = (float *)malloc(nx * sizeof(float));
   memset(ARMask, 0, nx*sizeof(float));
   BRMask  = (float *)malloc(nx * sizeof(float));
   memset(BRMask, 0, nx*sizeof(float));
   ABRMask = (float *)malloc(nx * sizeof(float));
   memset(ABRMask, 0, nx*sizeof(float));
   ARTotal  = (float *)malloc(nx * sizeof(float));
   memset(ARTotal, 0, nx*sizeof(float));
   BRTotal  = (float *)malloc(nx * sizeof(float));
   memset(BRTotal, 0, nx*sizeof(float));
   ABRTotal = (float *)malloc(nx * sizeof(float));
   memset(ABRTotal, 0, nx*sizeof(float));

/*
          now read in the real A data
*/
   
   idata = open_imaging_fdata_table(inFile, &ierr);
   if (ierr !=0) {
      printf("error %i opening imaging data in Afile %s\n", 
         ierr, AFile);
      return;
   }
   inData = make_imaging_fdata_row(idata);
   i = fits_get_num_rows (inFile,  &nRow, &ierr);
   ii = 1;
   i = read_imaging_fdata_table(inFile, idata, inData, ii++, &ierr);
   for (ix=0;ix<npix;ix++) ATarget[0][ix] = inData->data[chan1][ix];
   for (ix=0;ix<npix;ix++) ATarget[1][ix] = inData->data[chan1+1][ix];
   for (iBin = 0; iBin<nbin;iBin++) {
      i = read_imaging_fdata_table(inFile, idata, inData, ii++, &ierr);
      for (ix=0;ix<npix;ix++) ABin[0][iBin][ix] = inData->data[chan1][ix];
      for (ix=0;ix<npix;ix++) ABin[1][iBin][ix] = inData->data[chan1+1][ix];
   }
   i = close_imaging_fdata_table(inFile, idata, &ierr);
/*
      open BFile 
*/
   i = fits_open_file(&inFile, BFile, READONLY, &ierr) ;
   if (ierr != 0) {
      printf("cfitsio error %i opening B file %s\n", ierr, BFile);
      printf("%s \n",strerror(errno));
      return 1;
   }
   idata = open_imaging_fdata_table(inFile, &ierr);
   if (ierr !=0) {
      printf("error %i opening imaging data in Bfile %s\n", 
         ierr, BFile);
      return;
   }
   ii = 1;
   i = read_imaging_fdata_table(inFile, idata, inData, ii++, &ierr);
   for (ix=0;ix<npix;ix++) BTarget[0][ix] = inData->data[chan1][ix];
   for (ix=0;ix<npix;ix++) BTarget[1][ix] = inData->data[chan1+1][ix];
   for (iBin = 0; iBin<nbin;iBin++) {
      i = read_imaging_fdata_table(inFile, idata, inData, ii++, &ierr);
      for (ix=0;ix<npix;ix++) BBin[0][iBin][ix] = inData->data[chan1][ix];
      for (ix=0;ix<npix;ix++) BBin[1][iBin][ix] = inData->data[chan1+1][ix];
   }
   i = close_imaging_fdata_table(inFile, idata, &ierr);

   if (ierr == 0) printf("   Successfully read in A+B input data.\n");
   if (doAB) {
      i = fits_open_file(&inFile, ABFile, READONLY, &ierr) ;
      if (ierr != 0) {
         printf("cfitsio error %i opening AB file %s\n", ierr, ABFile);
         printf("%s \n",strerror(errno));
         return 1;
      }
      idata = open_imaging_fdata_table(inFile, &ierr);
      if (ierr !=0) {
         printf("error %i opening imaging data in ABfile %s\n", 
            ierr, ABFile);
      return;
      }
      ii = 1;
      i = read_imaging_fdata_table(inFile, idata, inData, ii++, &ierr);
      for (ix=0;ix<npix;ix++) ABTarget[0][ix] = inData->data[chan1][ix];
      for (ix=0;ix<npix;ix++) ABTarget[1][ix] = inData->data[chan1+1][ix];
      for (iBin = 0; iBin<nbin;iBin++) {
         i = read_imaging_fdata_table(inFile, idata, inData, ii++, &ierr);
         for (ix=0;ix<npix;ix++) ABgin[0][iBin][ix] = inData->data[chan1][ix];
         for (ix=0;ix<npix;ix++) ABgin[1][iBin][ix] = inData->data[chan1+1][ix];
      }
      i = close_imaging_fdata_table(inFile, idata, &ierr);
      if (ierr == 0) printf("   Successfully read in A*B input data.\n");
   }
/*   kill_imaging_fdata_row(inData); */
/*
      get mask array
*/
   oirGetMaskImages(maskFile, maskImages, DetectorDef->nregion, nxList, nyList, &ierr);
   if (ierr != 0) {
      printf("error %i reading mask data from file %s\n", ierr, maskFile);
      return 1;
   }
/*
      shift mask images
*/
   if (autoShift) maskShift = findMaskShift(maskImages[chan1], maskImages[1+chan1],
      ATarget[0], ATarget[1], BTarget[0], BTarget[1], nx, ny);
      printf("   Shifting masks vertically by %f pixel(s).\n", maskShift);
   if (maskShift !=0) shiftMaskImages(maskImages, maskShift, nRegion, nx, ny);
/*
 *    output shifty version of mask ?
 */
   if (outMaskFileName[0] != 0) putOutMask(maskImages, nRegion, nx, ny,
      outMaskFileName, DetectorDef, &ierr);

   if (ierr !=0) return 1;
/*
      calculate total fluxes 
*/
   for (ix = 0; ix <nx; ix++) {
      for (iy=yBottom; iy <=yTop; iy++) {
         ii = ix + nx*(long)iy;
         for (j=0; j<2; j++) {
            ATotal[ix] = ATotal[ix] + ATarget[j][ii] ;
            BTotal[ix] = BTotal[ix] + BTarget[j][ii];
            if (doAB) {
               xx = ATarget[j][ii] - BTarget[j][ii]; /* (A-B) */
               xx = xx * xx;  /* (A-B)^2 */
               xx = ABTarget[j][ii]*ABTarget[j][ii] - xx;/* (A+B)^2 - (A-B)^2 ~ 4AB */
               xx = 1.0 * xx;
            } else xx = ATarget[j][ii] * BTarget[j][ii];
            if (xx > 0) ABTotal[ix] = ABTotal[ix] + sqrt(xx);
         }
      } /* iy */
/*
      calculate rmss from binned totals
*/
      for (i=0; i<nbin; i++) {
         Ascratch  = 0.;
         Bscratch  = 0.;
         ABscratch = 0.;
         for (iy=yBottom; iy <=yTop; iy++) {
            ii = ix + nx*(long)iy;
            for (j=0; j<2; j++) {
               Ascratch = Ascratch + ABin[j][i][ii];
               Bscratch = Bscratch + BBin[j][i][ii];
               if (doAB) {
                  xx = ABin[j][i][ii] - BBin[j][i][ii]; /* (A-B) */
                  xx = xx * xx;  /* (A-B)^2 */
                  xx = ABgin[j][i][ii]*ABgin[j][i][ii] - xx;/* (A+B)^2 - (A-B)^2 ~ 4AB */
                  xx = 1.0 * xx;
               } else xx       = ABin[j][i][ii] * BBin[j][i][ii];
               if (xx > 0) ABscratch = ABscratch + sqrt(xx);
            }
         } /* iy */
         Ascratch     = Ascratch  - ATotal[ix];
         Bscratch     = Bscratch  - BTotal[ix];
         ABscratch    = ABscratch - ABTotal[ix];
         ARTotal[ix]  = ARTotal[ix]  + Ascratch*Ascratch;
         BRTotal[ix]  = BRTotal[ix]  + Bscratch*Bscratch;
         ABRTotal[ix] = ABRTotal[ix] + ABscratch*ABscratch;
      } /* iBin */
      if (ARTotal[ix] > 0)  ARTotal[ix]  = sqrt(ARTotal[ix]/(nbin*(nbin-1)));
      if (BRTotal[ix] > 0)  BRTotal[ix]  = sqrt(BRTotal[ix]/(nbin*(nbin-1)));
      if (ABRTotal[ix] > 0) ABRTotal[ix] = sqrt(ABRTotal[ix]/(nbin*(nbin-1)));
   } /* ix */

/*
      calculate masked fluxes
*/
   for (ix = 0; ix <nx; ix++) {
      for (iy=yBottom; iy <=yTop; iy++) { 
         ii = ix + nx*(long)iy;
         for (j=0; j<2; j++) {
            AMask[ix] = AMask[ix] + ATarget[j][ii] * maskImages[chan1+j][ii];
            BMask[ix] = BMask[ix] + BTarget[j][ii] * maskImages[chan1+j][ii];
            mm = maskImages[chan1+j][ii];
            if (doAB) {
               xx = ATarget[j][ii] - BTarget[j][ii]; /* (A-B) */
               xx = xx * xx;  /* (A-B)^2 */
               xx = ABTarget[j][ii]*ABTarget[j][ii] - xx;/* (A+B)^2 - (A-B)^2 ~ 4AB */
               xx = 1.0 * xx * mm * mm;
            } else xx = ATarget[j][ii]*BTarget[j][ii] * mm * mm;
            if (xx > 0) ABMask[ix] = ABMask[ix] + sqrt(xx);
         }
      } /* iy */
/*
      calculate rmss from binned totals
*/
      for (i=0; i<nbin; i++) {
         Ascratch  = 0.;
         Bscratch  = 0.;
         ABscratch = 0.;
         for (iy=yBottom; iy <=yTop; iy++) {
            ii = ix + nx*(long)iy;
            for (j=0;j<2;j++) {
               Ascratch = Ascratch + ABin[j][i][ii] * maskImages[chan1+j][ii];
               Bscratch = Bscratch + BBin[j][i][ii] * maskImages[chan1+j][ii];
               mm = maskImages[chan1+j][ii];
               if (doAB) {
                  xx = ABin[j][i][ii] - BBin[j][i][ii]; /* (A-B) */
                  xx = xx * xx;  /* (A-B)^2 */
                  xx = ABgin[j][i][ii]*ABgin[j][i][ii] - xx;/* (A+B)^2 - (A-B)^2 ~ 4AB */
                  xx = 1.0 * xx;
               }   else xx = ABin[j][i][ii]*BBin[j][i][ii] * mm * mm;
               if (xx > 0) ABscratch = ABscratch + sqrt(xx);
            }
         } /* iy */
         Ascratch  = Ascratch  - AMask[ix];
         Bscratch  = Bscratch  - BMask[ix];
         ABscratch = ABscratch - ABMask[ix];
         ARMask[ix]  = ARMask[ix]  + Ascratch*Ascratch;
         BRMask[ix]  = BRMask[ix]  + Bscratch*Bscratch;
         ABRMask[ix] = ABRMask[ix] + ABscratch*ABscratch;
      } /* iBin */
      if (ARMask[ix] > 0)  ARMask[ix]  = sqrt(ARMask[ix]/(nbin*(nbin-1)));
      if (BRMask[ix] > 0)  BRMask[ix]  = sqrt(BRMask[ix]/(nbin*(nbin-1)));
      if (ABRMask[ix] > 0) ABRMask[ix] = sqrt(ABRMask[ix]/(nbin*(nbin-1)));
   } /* ix */
   DetectorDef->nregion = 1;
   DetectorDef->naxis[0][1] = 1;
   iKey = table_find_key(DetectorDef->def, "NREGION");
   table_put_keyword(DetectorDef->def, iKey, (void *)&one, &ierr);

   odata = make_imaging_fdata_table_def (outFile, DetectorDef, 1, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error creating data definition\n");
      return 1;
   }
   outData = make_imaging_fdata_row(odata);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ATotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 1, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = BTotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 2, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ABTotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 3, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = AMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 4, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = BMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 5, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ABMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 6, &ierr);
   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ARTotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 7, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = BRTotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 8, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ABRTotal[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 9, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ARMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 10, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = BRMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 11, &ierr);

   for (ix=0;ix<nx;ix++) outData->data[0][ix] = ABRMask[ix];
   i = write_imaging_fdata_table(outFile, odata, outData, 12, &ierr);
   if (ierr == 0) printf("   Successfully wrote output data.\n");
   i = close_imaging_fdata_table(outFile, odata, &ierr);
   kill_imaging_fdata_row(outData);

   for (j=0;j<2;j++) {
      free(ATarget[j]);
      free(BTarget[j]);
      for (i=0; i<nbin; i++) {
         free(ABin[j][i]); 
         free(BBin[j][i]); 
         if (doAB) free(ABgin[j][i]);
      }
      free(ABin[j]); 
      free(BBin[j]); 
      if (doAB) free(ABgin[j]);
   }
   free(AMask);
   free(BMask);
   free(ABMask);
   free(ATotal);
   free(BTotal);
   free(ABTotal);
   free(ARMask);
   free(BRMask);
   free(ABRMask);
   free(ARTotal);
   free(BRTotal);
   free(ABRTotal);
   for (i=0;i<nRegion;i++) free(maskImages[i]);
   return 0;
}
